/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.annotation.valueprovider;

import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.utils.ComponentArchitectureUtils;
import org.fortiss.af3.exploration.moea.model.annotation.AnnotationPackage;
import org.fortiss.af3.exploration.moea.model.annotation.ComponentDiverseImplRef;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureValueProviderBase;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;

/**
 * {@link IAnnotationValueProvider} for referencing {@link Component}s that represent a diverse
 * implementation of a component that is part of a functional description of a
 * {@link ComponentArchitecture}
 * 
 * @author barner
 */
public class ComponentImplDiversityRefValueProvider extends
		EStructuralFeatureValueProviderBase<ComponentDiverseImplRef> {

	/**
	 * {@link EStructuralFeatureDescriptor} that filters the available objects to be selected for
	 * the reference provided by this annotation.
	 */
	private static class DiversityRefFilterEStructuralFeatureDescriptor extends
			EStructuralFeatureDescriptor {

		/** Constructs a new {@link DiversityRefFilterEStructuralFeatureDescriptor}. */
		public DiversityRefFilterEStructuralFeatureDescriptor(EReference eReference,
				EReferenceScope eReferenceScope) {
			super(eReference, eReferenceScope);
		}

		/** {@inheritDoc} */
		@Override
		public boolean isAvailableObject(EObject obj, IAnnotatedSpecification specification,
				EObject modelElement) {

			// Only offer only Components
			if(!(obj instanceof Component)) {
				return false;
			}
			Component refComponent = (Component)obj;

			// Allow references to Components implementing diversity only if the "abstract", i.e.
			// the referencing, Component is atomic.
			if(modelElement instanceof Component &&
					ComponentArchitectureUtils.isAtomicComponent((Component)modelElement)) {

				ComponentArchitecture targetCompArch =
						getParentElement(obj, ComponentArchitecture.class, false);

				// Check, whether the the parent ComponentArchitcture is the component pool (via the
				// name of the top-level Component).
				if(targetCompArch != null &&
						targetCompArch.getTopComponent().getName().contains("Component Pool")) {
					// Do not allow self references inside the component pool.
					ComponentArchitecture sourceCompArch =
							getParentElement(modelElement, ComponentArchitecture.class, false);
					if(sourceCompArch != null && sourceCompArch != targetCompArch) {
						// Only return diverse Component implementations if the referenced Component
						// is a sub-element of a Component in the Component Pool. (2nd layer)
						if(getDistToTopLevel(refComponent, targetCompArch) == 1) {
							return true;
						}
					}
				}
			}
			return false;
		}
	}

	/**
	 * Returns the "level" in the hierarchy of a "pure" {@link ComponentArchitecture} i.e. a
	 * {@link ComponentArchitecture} consisting only of {@link Component}s. The {@link Component}s
	 * contained in the top level component are considered to be at level 0.
	 */
	private static int getDistToTopLevel(Component comp, ComponentArchitecture compArch) {
		if(comp == compArch.getTopComponent()) {
			return -1;
		}

		Component parentComp = comp.getParentComponent();
		int level = 0;
		while(parentComp != null && parentComp != compArch.getTopComponent()) {
			parentComp = parentComp.getParentComponent();
			level++;
		}

		return level;
	}

	/** Constructs a value provider for the (derived) {@link ComponentDiverseImplRef} annotation. */
	public ComponentImplDiversityRefValueProvider() {
		super(AnnotationPackage.Literals.COMPONENT_DIVERSE_IMPL_REF,
				new DiversityRefFilterEStructuralFeatureDescriptor(
						AnnotationPackage.Literals.COMPONENT_DIVERSE_IMPL_REF__COMPONENT_REF,
						EStructuralFeatureDescriptor.EReferenceScope.PROJECT));
	}

	/** {@inheritDoc} */
	@Override
	public String getAnnotationName(ComponentDiverseImplRef specification) {
		return "Diverse Implementations";
	}
}
