/*-------------------------------------------------------------------------+
| Copyright 2015 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode;

import static org.fortiss.af3.exploration.util.ExplorationUtils.isDebugVerboseEnabled;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.create.ComposableCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.MissingDependencyException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DecoderDependencyGraph;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.opt4j.core.Genotype;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Decoder;

/**
 * Base class for decoding {@link CompositeGenotype}s. It is responsible for managing the decoding
 * the sub-{@link ComposableGenotype}s that form the {@link CompositeGenotype}.
 * 
 * @author diewald
 */
public abstract class CompositeDecoderBase<C extends CompositeGenotype<Class<? extends ComposableGenotype>, ComposableGenotype>, P extends Phenotype>
		implements CompositeDecoder<C, P> {
	/**
	 * References the {@link DecoderDependencyGraph} which defines execution orders and which
	 * {@link Phenotype}s are needed within the DSE.
	 */
	protected DecoderDependencyGraph depGraph;

	/**
	 * Collection of {@link Phenotype} types that are needed to generated the {#link
	 * requestedPhenotypeTypes}.
	 */
	Collection<Class<? extends Phenotype>> requiredPhenotypeTypes;

	/**
	 * Collection of {@link Phenotype}s that were externally demanded as a solution to
	 * input DSE problem.
	 */
	Collection<Class<? extends Phenotype>> requestedPhenotypeTypes;

	/** Relates decoders with the {@link Genotype}-Type that they decode. */
	private Map<ComposableDecoder<? extends ComposableGenotype, ? extends Phenotype>, Class<? extends ComposableGenotype>> registeredSubDecoders =
			new HashMap<>();

	/**
	 * {@link LinkedHashMap} referencing the required decoders for the given optimization problem
	 * that respects the execution order calculated by the {@link DecoderDependencyGraph}.
	 */
	private LinkedHashMap<ComposableDecoder<ComposableGenotype, Phenotype>, Class<? extends ComposableGenotype>> subDecoders =
			new LinkedHashMap<>();

	/**
	 * Flag to determine whether all required {@link ComposableDecoder}s (indicated by the
	 * {@link DecoderDependencyGraph}) are registered.
	 */
	private boolean isDecoderRegistrationChecked = false;

	/** Default constructor that requires a representation of the input models. */
	public CompositeDecoderBase(DecoderDependencyGraph depGraph) {
		this.depGraph = depGraph;
	}

	/**
	 * Decodes the given {@link Genotype} which is a {@link CompositeGenotype}. It launches all
	 * registered {@link Decoder}s for its {@link ComposableGenotype}s that compose the given
	 * {@link CompositeGenotype}.
	 * 
	 * This method is called from {@link CompositeDecoderBase#decode(CompositeGenotype)}. It is
	 * implemented separately such that type-to-object equivalence is enforced for the background
	 * maps.
	 * 
	 * @param genotype
	 *            {@link Genotype} (i.e., a {@link CompositeGenotype}) for which the sub-decoders
	 *            will be launched.
	 * @return The decoded {@link Phenotype} which contains the decoded {@link Phenotype}s.
	 * @throws DecodingException
	 *             if a required {@link ComposableDecoder} is not registered, or one of the
	 *             sub-Decoders produces an error.
	 * @throws MissingDependencyException
	 *             if some required {@link IExplorationEncoding} of a {@link ComposableDecoder} is
	 *             not provided by another one, a {@link ComposableCreator}.
	 */
	public <SP extends Genotype, CG extends ComposableGenotype, CP extends Phenotype> P
			decodeInternal(C genotype) throws DecodingException, MissingDependencyException {
		Map<Class<CP>, CP> subPhenotypes = new HashMap<>();
		FlatPhenotypeMap<Phenotype> decodedPhenotypes = new FlatPhenotypeMap<>();
		if(!isDecoderRegistrationChecked) {
			isDecoderRegistrationChecked = true;
			for(ComposableDecoder<?, ?> decoder : depGraph.getOperatorExecOrder()) {
				if(registeredSubDecoders.containsKey(decoder)) {
					@SuppressWarnings("unchecked") ComposableDecoder<ComposableGenotype, Phenotype> castedDecoder =
							(ComposableDecoder<ComposableGenotype, Phenotype>)decoder;
					subDecoders.put(castedDecoder, registeredSubDecoders.get(decoder));
				} else {
					throw new DecodingException(
							genotype,
							decoder,
							"The sub-decoder " +
									decoder.getClass().getSimpleName() +
									" has been found in the dependency graph but it has not been regirstered at this composite decoder.");
				}
			}
		}

		for(Entry<ComposableDecoder<ComposableGenotype, Phenotype>, Class<? extends ComposableGenotype>> decodeEntry : subDecoders
				.entrySet()) {
			@SuppressWarnings("unchecked") Class<CG> genotypeKey =
					(Class<CG>)decodeEntry.getValue();
			CG decodeableGenotype = genotype.get(genotypeKey);

			@SuppressWarnings("unchecked") ComposableDecoder<CG, CP> castedDecoder =
					(ComposableDecoder<CG, CP>)decodeEntry.getKey();
			@SuppressWarnings("unchecked") Class<CP> decType =
					(Class<CP>)castedDecoder.getOutputType();
			CP curPhenotype = null;

			// Decode a "normal" phenotype; Update a "Hybrid Phenotype" whose encoded information
			// shall not be just overridden.
			if(Genotype.class.isAssignableFrom(decType)) {
				CP oldGenotype = genotype.get(decType);
				curPhenotype =
						castedDecoder.updateComposable(decodeableGenotype, oldGenotype,
								decodedPhenotypes);
			} else {
				curPhenotype =
						castedDecoder.decodeComposable(decodeableGenotype, decodedPhenotypes);
			}

			// Catch erroneous phenotypes (failed decoding).
			if(curPhenotype != null) {
				// TODO #checking: invoke the checking here.

				@SuppressWarnings("unchecked") Class<CP> penotypeClass =
						(Class<CP>)curPhenotype.getPTypeToRegister();
				subPhenotypes.put(penotypeClass, curPhenotype);
				decodedPhenotypes.putEncoding(penotypeClass, curPhenotype);

				// Include all demanded inputs of a decoder that are Phenotypes as well, but which
				// are provided by the systemModelAdapter. They may be needed by the model
				// transformation framework, hence include them in the output encodings.
				for(Class<?> encodingClass : castedDecoder.getRequiredTypes()) {
					if(Phenotype.class.isAssignableFrom(encodingClass)) {
						@SuppressWarnings("unchecked") Class<? extends Phenotype> castedEncClass =
								(Class<? extends Phenotype>)encodingClass;
						decodedPhenotypes.putEncoding(castedEncClass,
								decodedPhenotypes.getEncoding(castedEncClass));
					}
				}

				if(curPhenotype instanceof ComposableGenotype) {
					addHybridGenoPhenotype(genotype, curPhenotype, penotypeClass);
				}
			}
		}

		P compositePhenotype = mergeSubPhenotypes(subPhenotypes);
		for(Entry<Class<CP>, CP> subPhenotype : subPhenotypes.entrySet()) {
			compositePhenotype.putSubPhenotype(subPhenotype.getKey(), subPhenotype.getValue());
		}

		return compositePhenotype;
	}

	/**
	 * <b>NOTE:</b> Sub-implementations MUST call {@code super.decode(genotype)}!!!
	 * <p>
	 * {@inheritDoc}
	 */
	@Override
	public P decode(C genotype) {
		try {
			return decodeInternal(genotype);
		} catch(DecodingException | MissingDependencyException e) {
			if(isDebugVerboseEnabled()) {
				e.printStackTrace();
			}
			throw new RuntimeException(e);
		}
	}

	/**
	 * Adds a hybrid geno-/phenotype to the given {@link CompositeGenotype}. A hybrid
	 * geno-/phenotype is a {@link ComposableGenotype} which is a Phenotype but also a Genotype
	 * that is used to calculate an dependent Phenotype.
	 */
	@SuppressWarnings("unchecked")
	private <V extends ComposableGenotype, M extends Phenotype> void addHybridGenoPhenotype(
			C genotype, M curPhenotype, Class<M> penotypeClass) {
		for(Class<? extends ComposableGenotype> registeredType : genotype.keySet()) {
			if(registeredType.isAssignableFrom(penotypeClass)) {
				genotype.put((Class<V>)penotypeClass, (V)curPhenotype);
				return;
			}
		}
		// Only relevant during development --> asserts should be enabled.
		assert (false) : "There is no decoder registered for the hybrid pheno- and genotype " +
				penotypeClass + ".";
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws IllegalArgumentException
	 */
	@SuppressWarnings("unchecked")
	@Override
	public <G extends ComposableGenotype, CP extends Phenotype, D extends ComposableDecoder<G, CP>>
			void registerSubDecoder(Class<G> subGenotype, D subDecoder)
					throws IllegalArgumentException {
		if(subGenotype == null || subDecoder == null) {
			return;
		}

		registeredSubDecoders.put(subDecoder,
				(Class<? extends ComposableGenotype>)subDecoder.getPrimaryInputType());

		ComposableDecoder<ComposableGenotype, Phenotype> castedSubDecoder =
				(ComposableDecoder<ComposableGenotype, Phenotype>)subDecoder;
		depGraph.addOperator(castedSubDecoder);
	}
}
