/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.opt4j.core.Genotype;
import org.opt4j.core.Objective;
import org.opt4j.core.problem.Creator;
import org.opt4j.core.problem.Decoder;
import org.opt4j.core.problem.Evaluator;
import org.opt4j.core.problem.ProblemModule;

import com.google.inject.TypeLiteral;
import com.google.inject.multibindings.Multibinder;

/**
 * Base class for a problem module that enables the registration of pre-constructed
 * creators/decoders/evaluators, and which thus can be constructed using a non-default constructor.
 * 
 * @author barner
 */
public abstract class ProblemModuleBase<G extends Genotype, R extends Creator<G>, D extends Decoder<G, ?>>
		extends org.opt4j.core.problem.ProblemModule {

	/** Stores the {@link Creator} bound to this {@link ProblemModule}. */
	protected R creator;

	/** Stores the {@link Decoder} bound to this {@link ProblemModule}. */
	protected D decoder;

	/** Stores the {@link Evaluator}s bound to this {@link ProblemModule}. */
	protected Collection<EvaluatorWrapper<?, ?, ?>> evaluators;

	/** Map that stores the associations between {@link ExplorationTarget}s and Objectives. */
	protected Map<ExplorationTarget<?>, Objective> targetObjectiveAssoc;

	/** Constructor. */
	public ProblemModuleBase() {
		evaluators = new HashSet<>();
		targetObjectiveAssoc = new ConcurrentHashMap<>();
	}

	/** Returns the {@link Objective} associated with the given {@link ExplorationTarget}. */
	public Objective getObjectiveOf(ExplorationTarget<?> target) {
		return targetObjectiveAssoc.get(target);
	}

	/**
	 * Registers a pre-constructed evaluator instances with the Opt4J framework.
	 */
	protected void addEvaluator(EvaluatorWrapper<?, ?, ?> evaluator) {
		assert (false);
		registerEvaluator(evaluator);
	}

	/**
	 * Registers the given {@link EvaluatorWrapper} with {@code this} class such that it can be used
	 * for the result validation.
	 */
	protected void registerEvaluator(EvaluatorWrapper<?, ?, ?> evaluator) {
		evaluators.add(evaluator);
	}

	/**
	 * Registers a {@link Collection} of pre-constructed evaluator instances with the Opt4J
	 * framework.
	 */
	protected void addEvaluators(Collection<EvaluatorWrapper<?, ?, ?>> evaluators) {

		// It is crucial that the Multibinder and the TypeLiteral are defined for
		// <Evaluator<Object>, since otherwise the binding of the evaluators will fail.
		Multibinder<Evaluator<Object>> multibinder =
				Multibinder.newSetBinder(binder(), new TypeLiteral<Evaluator<Object>>() { //
						});

		// Configure evaluators (based on provided goal specification)
		for(EvaluatorWrapper<?, ?, ?> eval : evaluators) {
			addEvaluator(eval, multibinder);
			evaluators.add(eval);
		}
	}

	/** Helper method to register a pre-constructed evaluator instance with the Opt4J framework. */
	@SuppressWarnings("unchecked")
	private void addEvaluator(EvaluatorWrapper<?, ?, ?> evaluator,
			Multibinder<Evaluator<Object>> multibinder) {

		// It is crucial that the Multibinder and the TypeLiteral are defined for
		// <Evaluator<Object>, since otherwise the binding of the evaluators will fail.

		// This is inspired from ProblemModule.addEvaluator(), which does not allow to pass
		// parameters to the constructor of the evaluators. Hence, evaluators are instantiated here,
		// and the binding is performed against these instances.
		multibinder.addBinding().toInstance((Evaluator<Object>)evaluator);
	}

	/** Returns the set of evaluators bound to this instance of the {@link ProblemModule}. */
	public Collection<EvaluatorWrapper<?, ?, ?>> getEvaluators() {
		return evaluators;
	}
}
