/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static java.lang.Math.pow;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedAcyclicTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.EvaluatorInjectorProvider;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.mappers.TaskMapper;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.time.DeadlineConstraint;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;

import com.google.inject.Provides;
import com.google.inject.assistedinject.Assisted;
import com.google.inject.name.Named;

/**
 * Evaluator for deadline objectives.
 * 
 * @author huang
 */
public class DeadlineConstraintEvaluator<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends ConstraintEvaluatorBase<Double, DeadlineConstraint, StrictTTSchedule<S, T>> {

	/**
	 * The source component which defines the component associated with the start time of this
	 * deadline.
	 */
	private Collection<ITaskAdapter<?>> sourceDeployableComponent;

	/**
	 * The target component which defines the component associated with the finish time of this
	 * deadline.
	 */
	private Collection<ITaskAdapter<?>> targetDeployableComponent;

	/** Stores the deadline for this goal. */
	private double deadline;

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public DeadlineConstraintEvaluator(Class<DeadlineConstraint> explorationTargetType)
			throws Exception {
		super(explorationTargetType);
	}

	/** Objective to track evaluation of this {@link ExplorationTarget}. */
	public DeadlineConstraintEvaluator(@Assisted DeadlineConstraint optGoal,
			TaskMapper<?> taskMapper) throws Exception {
		super(optGoal);
		TaskMapper<EObject> localTaskMapper = (TaskMapper<EObject>)taskMapper;
		sourceDeployableComponent = new ArrayList<>();
		targetDeployableComponent = new ArrayList<>();

		sourceDeployableComponent.add(localTaskMapper.getAdapter(optGoal.getStartComponent()));
		targetDeployableComponent.add(localTaskMapper.getAdapter(explorationTarget
				.getEndComponent()));

		deadline = explorationTarget.getDeadline();

		// TODO(#2960): Reenable the Deadline Constraint evaluator based on DSML expressions.
		// Currently, it is commented out to allow a reduction of the input code.

		// TODO: AD: The cast to the AF3 systemmodeladapter breaks the abstration the goals. It
		// would be rather required to define the source and sink components as
		// IDeployableComponents in the ecore model. The transformation from AF3 components to the
		// IDeployableCompontes should be done when the data from the GUI is written to the model.
		// ITaskAdapter<?> origStartComponent =
		// ((AF3SystemModelAdapter)systemModelAdapter).getTaskAdapter(explorationTarget
		// .getStartComponent());
		// if(origStartComponent.isInstantiation()) {
		// sourceDeployableComponent.add(origStartComponent);
		// } else {
		// sourceDeployableComponent.addAll(origStartComponent.getReplacementComponents());
		// }
		//
		// ITaskAdapter<?> origTargetComponent =
		// ((AF3SystemModelAdapter)systemModelAdapter).getTaskAdapter(explorationTarget
		// .getEndComponent());
		// if(origStartComponent.isInstantiation()) {
		// targetDeployableComponent.add(origTargetComponent);
		// } else {
		// targetDeployableComponent.addAll(origTargetComponent.getReplacementComponents());
		// }
	}

	/**
	 * evaluate the deadline objective
	 * if the deadline is met, we set the fitness value to -1
	 * if the deadline is violated, we set the fitness value to the gap between the actual latency
	 * and the deadline
	 */
	@Provides
	@Named("DeadlineConstraintEvaluator")
	public DoubleValue evaluateGoal(StrictTTSchedule<S, T> schedule,
			InstantiatedAcyclicTaskGraphEncoding<S> acycItgEnc,
			@Named("HyperPeriod") Double hyperPeriod) {
		double retval = 0;

		// One TaskGraph suffices as we know that "replacements" are always located in the same task
		// graph.
		TaskGraph taskGraph =
				acycItgEnc.getTaskGraphOf(sourceDeployableComponent.stream().findFirst().get());
		// we need to check each iteration in the hyper period
		for(int i = 0; i < hyperPeriod / taskGraph.getPeriod(); i++) {
			// Required for the lambda operators.
			final int j = i;

			// the start time is the release time of this iteration, which is the beginning
			// of the period
			// double start = i * tg.getPeriod();
			ITaskAdapter<?> earliestComp =
					sourceDeployableComponent.stream()
							.min(Comparator.comparing(t -> schedule.getFirstStartTime(t, j))).get();
			double start = schedule.getFirstStartTime(earliestComp, i);
			ITaskAdapter<?> latestComp =
					targetDeployableComponent.stream()
							.min(Comparator.comparing(t -> schedule.getLastEndTime(t, j))).get();
			double end = schedule.getLastEndTime(latestComp, i);
			double latency = end - start;

			// TODO: fix the logic below, it is a stylistic nightmare.
			if(latency <= 0)
				continue;

			// see if the deadline is met
			if(!(latency > deadline)) {
				// deadline ok, keep the retval unchanged
			} else {
				// deadline violated
				if(retval < 0) {
					retval = latency - deadline;
				} else {
					retval += latency - deadline;
				}
			}
		}

		return new DoubleValue(retval > 0 ? pow(retval * getIterationCountAndIncrement(), 2) : 0);
	}

	/** {@inheritDoc} */
	@Override
	public boolean validate(ExplorationConstraint<?> expTarget,
			EvaluatorInjectorProvider evalInjProvider, Value<?> value) {
		// TODO: Readd the getLimitation support or find another way to express the same.
		Double evalValue = value.getDouble();
		if(evalValue > ((DeadlineConstraint)expTarget).getLimitation() - NUMERIC_IMPRECISION) {
			return false;
		}
		return true;
	}
}
