/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.module;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.CompositeEvaluator2;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.exception.ExplorationRuntimeException;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.dsl_v2.model.expression.IExpression;
import org.fortiss.af3.exploration.dsl_v2.model.function.Maximize;
import org.fortiss.af3.exploration.model.ExplorationObjective;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.smt.model.SMTObjective;
import org.opt4j.core.Objective;
import org.opt4j.core.Objective.Sign;
import org.opt4j.core.problem.Evaluator;

/**
 * {@link Evaluator} module whose dependencies are defined using guice. Used to define sub-problems
 * of a DSE process.
 * 
 * @author diewald
 */
public abstract class EvaluatorModule<R, T extends ExplorationTarget<R>, P extends Phenotype>
		extends CompositeEvaluator2<R, P> implements IEvaluatorModule<R, T, P> {
	/**
	 * {@link ExplorationTarget} type to which the {@link EvaluatorModule} applies. Used for
	 * registering the EvaluatorModule with the {@link IExplorationEvaluatationService}.
	 */
	protected Class<? extends T> explorationTargetType;

	/**
	 * References the {@link ExplorationTarget} instance when a new Evaluator is created from the
	 * {@link IExplorationEvaluatationService}.
	 */
	protected T explorationTarget;

	/** PROTOTYPE Constructor. */
	public EvaluatorModule(Class<? extends T> explorationTargetType) {
		this.explorationTargetType = explorationTargetType;
	}

	/** Standard constructor for evaluator instances. */
	@SuppressWarnings("unchecked")
	public EvaluatorModule(T explorationTarget) {
		this((Class<T>)explorationTarget.getClass());
		this.explorationTarget = explorationTarget;

		String objectiveName =
				explorationTarget.getName() == null ? explorationTarget.getClass().toString() +
						explorationTarget.getId() : explorationTarget.getName() +
						explorationTarget.getId();

		Sign optDirection = Sign.MIN;
		if(explorationTarget instanceof ExplorationObjective) {
			// TODO: Reconsider the wrapping / passing to the ExpressionEvaluators. Probably, the
			// evaluators only need to be aware of the systemModelAdapter and the exploration
			// expression.
			IExpression expr = explorationTarget.getExpression();
			if(expr instanceof Maximize) {
				optDirection = Sign.MAX;
			}
		}

		objective = new Objective(objectiveName, optDirection);
	}

	/** Standard constructor for evaluator instances. */
	@SuppressWarnings("unchecked")
	public EvaluatorModule(SMTObjective explorationTarget) {
		this((Class<T>)explorationTarget.getClass());
		this.explorationTarget = (T)explorationTarget;

		String objectiveName =
				explorationTarget.getName() == null ? explorationTarget.getClass().toString() +
						explorationTarget.getId() : explorationTarget.getName() +
						explorationTarget.getId();
		// TODO: Determine / define the optimization direct by the expression / or by a passed
		// parameter.
		objective = new Objective(objectiveName, Sign.MIN);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<T> getExplorationTargetType() {
		return (Class<T>)explorationTargetType;
	}

	/** {@inheritDoc} */
	@Override
	public T getExplorationTarget() {
		return explorationTarget;
	}

	/** {@inheritDoc} */
	@Override
	public int getPriority() {
		// TODO: Return the priority specified in the ExplorationTarget.
		return 0;
	}

	/** {@inheritDoc} */
	@Override
	@SuppressWarnings("unchecked")
	protected Set<IEvaluatorModule<R, ?, ?>> getSubEvaluators() {
		return (Set<IEvaluatorModule<R, ?, ?>>)(Set<?>)subEvaluators;
	}

	/** {@inheritDoc} */
	@Override
	public <M extends IEvaluatorModule<R, T, P>> M newInstance(T explorationTarget) {
		try {
			@SuppressWarnings("unchecked") Constructor<M> moduleInstConstructor =
					(Constructor<M>)getClass().getConstructor(explorationTargetType);
			return moduleInstConstructor.newInstance(explorationTarget);
		} catch(NoSuchMethodException nsme) {
			throw new ExplorationRuntimeException(
					"Could not find a constructor in the EvaluatorModule " +
							getClass().getSimpleName() +
							" that takes the ExplorationTarget " +
							explorationTargetType.getSimpleName() +
							" as an input parameter. This is required to create EvaluationModules for the given set of ExplorationTargets.",
					nsme);
		} catch(SecurityException | InstantiationException | IllegalAccessException |
				IllegalArgumentException | InvocationTargetException stdE) {
			throw new ExplorationRuntimeException(
					"Could not create an Instance of the EvaluatorModule " +
							getClass().getSimpleName() + ".", stdE);
		}
	}
}
