/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.comm;

import java.util.Collection;
import java.util.HashSet;

import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

/**
 * Unidirectional message between a sender and a receiver (identified by their respective
 * {@link TimeSlot} of the {@link StrictTTSchedule}). While data dependencies between
 * deployable components mapped to the same deployment target are assumed to be implemented using
 * shared memory buffers (and the communication to consume no time), data dependencies between
 * deployable components mapped to different deployment targets are modeled using explicit messages
 * via a communication resource (e.g. a TDMA arbitrated bus).
 * 
 * <ul>
 * <li>TODO: Consider memory access times for intra-deployment unit communication</li>
 * <li>TODO: Replace {@link Message} with {@link Channel} from the logical architecture, or an
 * according adapter class?</li>
 * </ul>
 * 
 * @author huang
 */
public class Message implements Comparable<Message> {

	/** References the sender of this {@link Message}, i.e. a deployed component/task. */
	private InstantiatedTaskMappingEntry sender;

	/** References the {@link IMappingEntry}s that are receivers of this {@link Message}. */
	private Collection<InstantiatedTaskMappingEntry> receiverMappingEntries;

	/** References the receiver of this {@link Message}, i.e. a deployed component/task. */
	private Collection<IResourceAdapter<?>> receiverResources;

	/**
	 * References the route (of platform resources) which this message must traverse to provide its
	 * content.
	 */
	private DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge> route;

	/** {@link ISignalAdapter} which is the logical counterpart of this {@link Message} */
	private ISignalAdapter<?, ?> signalAdapter;

	/** {@link TimeSlot} in which sender of message is scheduled */
	private TimeSlot senderTimeSlot;

	/** {@link TimeSlot} in which receiver of message is scheduled */
	private HashSet<TimeSlot> receiverTimeSlots;

	/** Worst-case transmission time of message on its {@link ITransmissionUnitAdapter} */
	private double wctt;

	/** Contains the calculated size of the message. */
	private long sizeInBytes;

	/**
	 * Constructs a new {@link Message} between a sender and a receiver (identified by their
	 * respective {@link IMappingEntry}s in the {@link IMappingEncoding}).
	 */
	public Message(ISignalAdapter<?, ?> signalAdapter, InstantiatedTaskMappingEntry deployedSender,
			Collection<? extends InstantiatedTaskMappingEntry> receiverEntries) {
		this.signalAdapter = signalAdapter;
		this.sender = deployedSender;
		sizeInBytes = signalAdapter.getSignalSize();

		this.receiverResources = new HashSet<IResourceAdapter<?>>();
		this.receiverMappingEntries = new HashSet<InstantiatedTaskMappingEntry>();
		for(InstantiatedTaskMappingEntry receiver : receiverEntries) {
			addReceiver(receiver);
		}

		this.receiverTimeSlots = new HashSet<TimeSlot>();
	}

	/**
	 * Constructs a new {@link Message} between a sender and a receiver (identified by their
	 * respective {@link TimeSlot}s in the {@link StrictTTSchedule}).
	 */
	public Message(ISignalAdapter<?, ?> channelAdapter, TimeSlot senderTimeSlot,
			HashSet<TimeSlot> receiverTimeSlots) {
		this(channelAdapter, senderTimeSlot, receiverTimeSlots, channelAdapter.getSignalSize());
	}

	/**
	 * Constructs a new {@link Message} between a sender and a receiver (identified by their
	 * respective {@link TimeSlot}s in the {@link StrictTTSchedule}).
	 */
	public Message(ISignalAdapter<?, ?> channelAdapter, TimeSlot senderTimeSlot,
			HashSet<TimeSlot> receiverTimeSlots, long messageSizeInBytes) {
		this.signalAdapter = channelAdapter;
		this.senderTimeSlot = senderTimeSlot;
		this.receiverTimeSlots = receiverTimeSlots;

		wctt = 0.0;
		sizeInBytes = messageSizeInBytes;

		this.receiverResources = new HashSet<IResourceAdapter<?>>();
		for(TimeSlot currentReceiverSlot : receiverTimeSlots) {
			addReceiverResource(currentReceiverSlot.getResource());
		}
	}

	/**
	 * Adds a sender time slot to this message, it the time.
	 */
	public boolean addMessageSenderTimeSlot(TimeSlot senderSlot) {
		if(senderSlot == null || senderSlot.getResource() == null ||
				senderSlot.getResource() != sender.getTarget()) {
			return false;
		}
		this.senderTimeSlot = senderSlot;
		return true;
	}

	/**
	 * Adds a receiver time slot to the receiver list of this message and calculates the route to
	 * the {@link IResourceAdapter} it has been deployed upon.
	 */
	public boolean addMessageReceiverTimeSlots(TimeSlot receiverSlot) {
		assert (receiverSlot != null) : "A reciever slot may not be null.";
		return receiverTimeSlots.add(receiverSlot);
	}

	/** Adds an {@link IMappingEntry} as a receiver of this message. */
	public void addReceiver(InstantiatedTaskMappingEntry receiver) {
		receiverMappingEntries.add(receiver);
		receiverResources.add(receiver.getTarget());
	}

	/** Adds an {@link IMappingEntry} as a receiver of this message. */
	public void addReceiverResource(IResourceAdapter<?> reciever) {
		receiverResources.add(reciever);
	}

	/** Returns the {@link ISignalAdapter} corresponding to this messages */
	public ISignalAdapter<?, ?> getSignalAdapter() {
		return signalAdapter;
	}

	/**
	 * Returns the size of the message (in bytes) which is defined via the data type and its
	 * implementation in the sending execution unit.
	 */
	public long getMessageSize() {
		return sizeInBytes;
	}

	/** Returns the {@link IMappingEntry} of the sender of this message. */
	public IMappingEntry<?, ?> getSender() {
		return sender;
	}

	/** Returns the {@link IMappingEntry}s that receive this message. */
	public Collection<InstantiatedTaskMappingEntry> getReceiverMappingEntries() {
		return receiverMappingEntries;
	}

	/**
	 * Returns the {@link Collection} of {@link IMappingEntry}s that are the receivers of this
	 * message.
	 */
	public Collection<IResourceAdapter<?>> getReceiverResources() {
		return receiverResources;
		// Collection<IResourceAdapter<?>> recvResources = new HashSet<>();
		// for(IMappingEntry entry : receiverMappingEntries) {
		// recvResources.add(entry.getTarget());
		// }
		// return recvResources;
	}

	/** Returns the route which this {@link Message} must traverse. */
	public DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge> getRoute() {
		return route;
	}

	/**
	 * Sets the route in the form of a {@link DefaultDirectedGraph} which this {@link Message} must
	 * traverse.
	 */
	public void setRoute(DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge> messageRoute) {
		route = messageRoute;
	}

	/** Returns the {@link TimeSlot} of the sender. */
	public TimeSlot getSenderTimeSlot() {
		return senderTimeSlot;
	}

	/** Returns the {@link TimeSlot} of the receiver. */
	public HashSet<TimeSlot> getReceiverTimeSlots() {
		return receiverTimeSlots;
	}

	/** Returns worst-case transmission time (in seconds) */
	public double getWCTT() {
		return wctt;
	}

	/** Sets the WCTT (in seconds) for this message. This method is used for message scheduling. */
	public void setWCTT(double calculatedWctt) {
		wctt = calculatedWctt;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String receiverStr = "[";
		for(TimeSlot currentReceiverSlot : receiverTimeSlots) {
			receiverStr += currentReceiverSlot + ", ";
		}
		return senderTimeSlot + "->" + receiverStr + "]";
	}

	/** {@inheritDoc} */
	@Override
	public int compareTo(Message compareMessage) {
		assert (this.senderTimeSlot != null);
		if(this.senderTimeSlot == null) {
			return -1;
		}
		assert (compareMessage.getSenderTimeSlot() != null);

		if(this.senderTimeSlot.getEndTime() > compareMessage.getSenderTimeSlot().getEndTime()) {
			return 1;
		} else if(this.senderTimeSlot.getEndTime() == compareMessage.getSenderTimeSlot()
				.getEndTime()) {
			return 0;
		} else {
			return -1;
		}
	}
}
