/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.castToRawColl;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExplorationRepairService;
import org.fortiss.af3.exploration.alg.service.IRepairModule;
import org.fortiss.tooling.kernel.service.base.ObjectAwareServiceBase;
import org.opt4j.core.Genotype;

/**
 * Implementation of the {@link IExplorationRepairService}.
 * It constructs and returns {@link IRepairModule} instances when the relveant fodules for a given
 * {@link Genotype} are queried.
 * 
 * @author diewald
 */
public class ExplorationRepairService extends ObjectAwareServiceBase<IRepairModule<?>>
		implements IExplorationRepairService {

	/** The evaluator provider extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.af3.exploration.alg.repairModule";

	/** The evaluator provider configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "repairModule";

	/** The evaluator class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "module";

	/** Holds the instance of the evaluation service. */
	private volatile static ExplorationRepairService INSTANCE;

	/** Constructor. */
	private ExplorationRepairService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	public static synchronized ExplorationRepairService getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationRepairService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		// TODO Auto-generated method stub
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		// TODO Auto-generated method stub
		return null;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <T extends Genotype> List<IRepairModule<T>> getRepairFunctionsOf(T genotype)
			throws ExplorationException {
		List<IRepairModule<T>> repairModuleInstances = new ArrayList<>();
		Collection<IRepairModule<?>> registeredHandlers =
				castToRawColl(getRegisteredHandlers(genotype.getClass()));

		for(IRepairModule<?> curModuleTemplate : registeredHandlers) {
			IRepairModule<T> moduleInstance;
			try {
				Constructor<IRepairModule<T>> extConstructor =
						(Constructor<IRepairModule<T>>)curModuleTemplate.getClass()
								.getConstructor(SystemModelAdapter.class);
				moduleInstance = extConstructor.newInstance();
			} catch(Exception e) {
				throw new ExplorationException(
						"Could not instantiate the repair module from its registered template. Please check if the required Constructors are present.",
						e);
			}

			repairModuleInstances.add(moduleInstance);
		}

		return repairModuleInstances;
	}
}
