/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.fortiss.af3.exploration.alg.service.ExplorationServiceManager.getService;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.ui.services.IEvaluationService;
import org.fortiss.af3.exploration.alg.ExplorationAlgActivator;
import org.fortiss.af3.exploration.alg.dse.module.IEvaluatorModule;
import org.fortiss.af3.exploration.alg.feature.IExplorationFeature;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationFeatureService;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.tooling.kernel.utils.LoggingUtils;

import com.google.inject.Injector;

/**
 * Utility methods for the interaction with the {@link IExplorationFeatureService}.
 * 
 * @author diewald
 */
public class ExplorationFeatureUtils {

	/**
	 * From a set of enabled {@link IExplorationFeature}s of the given {@link DseSpecification},
	 * this method returns those features that are executable w.r.t. the set of available data from
	 * the input models provided by the {@code dseInjector}.
	 * 
	 * @param dseSpec
	 *            contains the enabled {@link IExplorationFeature}s.
	 * @param dseInjector
	 *            provides the available input models.
	 * @param logDisabled
	 *            if non-applicable shall be logged.
	 * @return collection of enabled and applicable {@link IExplorationFeature}s.
	 */
	public static Collection<IExplorationFeature> getEnabledApplicableFeatures(
			DseSpecification dseSpec, Injector dseInjector, boolean logDisabled) {
		Collection<IExplorationFeature> userEnabledFeatures =
				getService(IExplorationFeatureService.class).getExplorationFeatures(
						dseSpec.getFeatures());
		Collection<IExplorationFeature> enabledFeatures = new HashSet<>();
		for(IExplorationFeature feature : userEnabledFeatures) {
			if(feature.isApplicableToInputs(dseInjector)) {
				enabledFeatures.add(feature);
			} else {
				// @CodeFormatterOff
				LoggingUtils.warning(ExplorationAlgActivator.getDefault(),
						"The exploration feature " + feature.getName() + " has been disabled" +
						" since the input models do not contain sufficient data to execute this" +
						" feature.");
				// @CodeFormatterOn
			}
		}
		return enabledFeatures;
	}

	/**
	 * Returns the set of enabled Evaluators from the {@link IEvaluationService}, subject to the
	 * given {@link ExplorationSpecification}.
	 * 
	 * @param expSpec
	 *            contains the {@link ExplorationTarget}s to evaluate.
	 * @return see above.
	 */
	public static Set<IEvaluatorModule<?, ?, ?>> getEnabledEvaluators(
			ExplorationSpecification expSpec) {
		return getService(IExplorationEvaluatationService.class).getEvaluatorInstances(
				new HashSet<>(expSpec.getTargets()), createPriorityComparator());
	}

	/** Comparator that prefers evaluators of {@link ExplorationConstraint}s. */
	private static Comparator<IEvaluatorModule<?, ?, ?>> createPriorityComparator() {
		return new Comparator<IEvaluatorModule<?, ?, ?>>() {
			@Override
			public int compare(IEvaluatorModule<?, ?, ?> o1, IEvaluatorModule<?, ?, ?> o2) {
				if((o1.getExplorationTarget() instanceof ExplorationConstraint<?>) &&
						(o2.getExplorationTarget() instanceof ExplorationConstraint<?>)) {
					return o1.hashCode() - o2.hashCode();
				} else if(o1.getExplorationTarget() instanceof ExplorationConstraint<?>) {
					return 1;
				} else if(o2.getExplorationTarget() instanceof ExplorationConstraint<?>) {
					return -1;
				}
				return o1.hashCode() - o2.hashCode();
			}
		};
	}
}
